/* Copyright Izak Burger 2005.
 * Licensed under the GPL. */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#define PAM_SM_SESSION
/* #define PAM_SM_AUTH */

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define LOCAL_LOG_PREFIX "PAM-script: "

static void _pam_log(int err, const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vsyslog(LOG_AUTH | err, format, args);
	va_end(args);
}

static int session_script(pam_handle_t *pamh, int flags, int argc, 
		const char **argv, const char *execname)
{
	char *user = NULL;
	char *realuser = NULL;
	char mybuf[256],myexec[256],myuser[16],cmd[256];
	int i;
	struct passwd *pw;
	pid_t pid;
	sigset_t newmask, oldmask;
	int retcode = 0;

	/* Block SIGCHLD */
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGCHLD);
	sigprocmask(SIG_BLOCK, &newmask, &oldmask);

	/* Zero myexec and myuser */
	memset(myexec,'\0',sizeof(myexec));
	memset(myuser,'\0',sizeof(myuser));

	/* Handle arguments */
	for (i=0; i < argc; i++) {
		char *junk;
		memset(mybuf,'\0',sizeof(mybuf));
		junk = strchr(argv[i], '=');
		/* If the option is too long or it contains no '=', it must be bad */
		if ((junk == NULL) || (junk - argv[i]) >= sizeof(mybuf)) {
			_pam_log(LOG_ERR, "Bad option: \"%s\"",
			    argv[i]);
			continue; /* next argument */
		}
		strncpy(mybuf, argv[i], MIN(junk - argv[i], sizeof(mybuf) - 1));
		if (strcmp(mybuf, execname)==0) {
			strncpy(myexec, junk + 1, sizeof(myexec) - 1);
			_pam_log(LOG_DEBUG,LOCAL_LOG_PREFIX "%s is: \"%s\"", execname, myexec);
		} else if (strcmp(mybuf, "runas")==0) {
			strncpy(myuser, junk + 1, sizeof(myuser) - 1);
			user = myuser;
		}
	}
	/* Get the username of the person logging in right now. */
	pam_get_item(pamh, PAM_USER, (void*)&realuser);
	if(realuser==NULL)
		return PAM_SESSION_ERR;

	/* If no user to run as was passed in the arguments, use current user */
	if (user==NULL)
		user = realuser;

	/* At this point we know user!=NULL.  Attempt to retrieve password info */
	pw = getpwnam(user);
	/* Unable to get password info, return session error */
	if (pw==NULL)
		return PAM_SESSION_ERR;
	_pam_log(LOG_DEBUG, LOCAL_LOG_PREFIX "Executing user is: \"%s\"", user);

	/* What command we are to be runnink? */
	if(myexec[0]) { /* User provided a command to run */
		strncpy(cmd, myexec, sizeof(cmd));
	} else {
		snprintf(cmd, sizeof(cmd), "/etc/security/%s", execname);
	}

	/* Time to fork, we check for existence and executability of the script later, because we
	 * need to do that as the user executing the script, as it is possible that the script is
	 * executable by root but not by the user */
	pid = fork();
	if(pid<0){
		_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unable to fork");
		return PAM_SESSION_ERR;
	} else if(pid==0) {
		/* ======== Child ========= */
		/* Order of setregid and setreuid is important, first group, then user */
		if (setregid(pw->pw_gid, pw->pw_gid)==0 && setreuid(pw->pw_uid, pw->pw_uid)==0){
			struct stat st;
			char *arge[4];
			if (stat(cmd, &st) || !(S_ISREG(st.st_mode)) || !(st.st_mode&(S_IXUSR|S_IXGRP|S_IXOTH))) {
				_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "%s does not exist or is not executable", cmd);
				/* If the arguments supplied us with a script to run and it is not runable, return
				 * non-zero as this is an error.  Otherwise, return zero since it simply means the
				 * default doesn't exist, ie it is unconfigured and that is okay in our book */
				exit(myexec[0]?1:0);
			}
			arge[0] = cmd;
			arge[1] = realuser;
			pam_get_item(pamh, PAM_SERVICE, (void*)&arge[2]); /* Pass service as second argument (thanks Hanno!) */
			arge[3] = '\0';
			execvp(arge[0], arge);
			/* If we reach this, execvp failed.  Exit with an error */
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "execvp error: %d", errno);
			exit(1);
		} else {
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unable to run as %s", user);
		}
		/* ============= Child Ends ======== */
	}
	/* if pid <=0, we would not be here, so pid>0, which means we are the parent */
	/* wait for the child to return */
	while(waitpid(pid, &retcode, 0)==-1){
		if(errno != EINTR){
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "waitpid error: %d", errno);
			return PAM_SESSION_ERR;
		}
	}
	if (!WIFEXITED(retcode)) {/* terminated abnormlly */
		_pam_log(LOG_ERR,LOCAL_LOG_PREFIX "Executing file: \"%s\" as user %s returned abnormally", cmd, user);
		return PAM_SESSION_ERR;
	}
	if(WEXITSTATUS(retcode)!=0){ /* Terminated normally, but non-zero */
		_pam_log(LOG_ERR,LOCAL_LOG_PREFIX "Executing file: \"%s\" as user %s returned %d", cmd, user, WEXITSTATUS(retcode));
		return PAM_SESSION_ERR;
	}

	/* Restore signal mask */
	sigprocmask(SIG_SETMASK, &oldmask, NULL);

	return PAM_SUCCESS;
}

/* Session management stuff */
PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	return session_script(pamh, flags, argc, argv, "onsessionopen");
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	return session_script(pamh, flags, argc, argv, "onsessionclose");
}
